import os
import unittest
import pandas as pd
import sqlite3
import shutil
import zipfile
import hashlib
import subprocess
import signal

from sourmash import load_file_as_signatures, save_signatures
from mashpit.build import create_connection
from mashpit.build import create_database
from mashpit.build import download_metadata
from mashpit.build import calculate_centroid
from mashpit.build import download_and_sketch_assembly
from mashpit.update import compare_tables


class TestCreateConnection(unittest.TestCase):
    def test_connection_is_not_none(self):
        conn = create_connection(":memory:")
        self.assertIsNotNone(conn)

    def test_connection_is_sqlite_connection(self):
        conn = create_connection(":memory:")
        self.assertIsInstance(conn, sqlite3.Connection)


class TestCreateDatabase(unittest.TestCase):
    def setUp(self):
        # Create an in-memory database and connect to it before each test
        self.conn = sqlite3.connect(":memory:")

    def tearDown(self):
        # Close the connection and destroy the database after each test
        self.conn.close()

    def test_metadata_table_exists(self):
        create_database(self.conn)
        c = self.conn.cursor()
        c.execute(
            "SELECT name FROM sqlite_master WHERE type='table' AND name='METADATA';"
        )
        result = c.fetchone()
        self.assertIsNotNone(result)

    def test_description_table_exists(self):
        create_database(self.conn)
        c = self.conn.cursor()
        c.execute("SELECT name FROM sqlite_master WHERE type='table' AND name='DESC';")
        result = c.fetchone()
        self.assertIsNotNone(result)

    def test_metadata_table_structure(self):
        create_database(self.conn)
        c = self.conn.cursor()
        c.execute("PRAGMA table_info(METADATA);")
        columns = [row[1] for row in c.fetchall()]
        expected_columns = [
            "biosample_acc",
            "taxid",
            "strain",
            "collected_by",
            "collection_date",
            "geo_loc_name",
            "isolation_source",
            "lat_lon",
            "serovar",
            "sub_species",
            "species",
            "genus",
            "host",
            "host_disease",
            "outbreak",
            "srr",
            "PDT_acc",
            "PDS_acc",
            "asm_acc",
        ]
        self.assertListEqual(columns, expected_columns)

    def test_description_table_structure(self):
        create_database(self.conn)
        c = self.conn.cursor()
        c.execute("PRAGMA table_info(DESC);")
        columns = [row[1] for row in c.fetchall()]
        expected_columns = ["name", "value"]
        self.assertListEqual(columns, expected_columns)


class TestDownloadMetadata(unittest.TestCase):
    def setUp(self):
        self.pathogen_name = "Kluyvera_intermedia"
        self.pd_version = None
        os.mkdir("tmp")
        self.tmpfolder = "tmp"

    def tearDown(self):
        shutil.rmtree(self.tmpfolder)

    def test_download_metadata(self):
        try:
            download_metadata(self.pathogen_name, self.pd_version, self.tmpfolder)
        except Exception as e:
            self.fail(f"download_metadata raised an exception: {e}")


class TestCalculateCentroid(unittest.TestCase):
    def setUp(self):
        # Define test inputs and expected output
        self.metadata_file = "test.metadata.tsv"
        self.distance_file = "test.reference_target.SNP_distances.tsv"
        self.expected_output = pd.DataFrame(
            {
                "PDS_acc": [
                    "PDS000010723.7",
                    "PDS000053359.1",
                    "PDS000049646.3",
                    "PDS000098715.1",
                ],
                "target_acc": [
                    "PDT000136707.1",
                    "PDT000627411.1",
                    "PDT000849087.1",
                    "PDT000122974.1",
                ],
            }
        )
        # sort the expected output by PDS_acc
        self.expected_output.sort_values(by=["PDS_acc"], inplace=True)
        self.tmpfolder = "tmp"
        os.mkdir(self.tmpfolder)
        # Copy test inputs to temporary directory
        shutil.copy(self.metadata_file, self.tmpfolder)
        shutil.copy(self.distance_file, self.tmpfolder)

    def tearDown(self):
        # Remove temporary directory
        shutil.rmtree(self.tmpfolder)

    def test_calculate_centroid_returns_expected_output(self):
        # Call calculate_centroid with test inputs
        df_metadata = pd.read_csv(self.metadata_file, sep="\t")
        calculate_centroid(
            df_metadata[~df_metadata["asm_acc"].isnull()],
            "test",
            self.tmpfolder,
        )
        # Compare actual and expected output
        actual_output = pd.read_csv(
            os.path.join(self.tmpfolder, "test_cluster_center.tsv"), sep="\t"
        )
        # sort the actual output by PDS_acc
        actual_output.sort_values(by=["PDS_acc"], inplace=True)
        pd.testing.assert_frame_equal(actual_output, self.expected_output)


class TestDownloadAndSketchAssembly(unittest.TestCase):
    def setUp(self):
        # Create a temporary directory to store downloaded and generated files
        self.hash_number = 1000
        self.kmer_size = 31
        self.gca_acc_list = ["GCA_015929625.1", "GCA_009649915.1", "GCA_001598315.1"]
        self.tmpfolder = "tmp"
        os.mkdir(self.tmpfolder)

    def tearDown(self):
        shutil.rmtree(self.tmpfolder)

    def test_download_and_sketch_assembly(self):
        download_and_sketch_assembly(
            self.gca_acc_list, self.hash_number, self.kmer_size, self.tmpfolder
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join(self.tmpfolder, "signature", "GCA_015929625.1.sig")
            )
        )


class TestBuildTaxonAndQuery(unittest.TestCase):
    def setUp(self):
        self.pathogen_name = "Listeria_innocua"
        self.pd_version = "PDG000000091.9"

    def tearDown(self):
        shutil.rmtree("test_listeria_innocua")
        shutil.rmtree("ncbi_dataset")
        # remove all files starting with GCA_022617975
        for file in os.listdir():
            if file.startswith("GCA_022617975"):
                os.remove(file)
        os.remove("ncbi_dataset.zip")
        for file in os.listdir():
            if file.endswith(".log"):
                os.remove(file)

    def test_build_taxonomy(self):

        subprocess.run(
            [
                "mashpit",
                "build",
                "taxon",
                "test_listeria_innocua",
                "--species",
                self.pathogen_name,
                "--pd_version",
                self.pd_version,
            ]
        )
        expected_sqlite_sha = (
            "f361ea7b23ded17f69f9e128e16e3ea7cc8598845979bf72c9e07844f65a38d9"
        )
        expected_signature_sha = (
            "4dc90abc2935a70fc615ac0111a8a27fc506b25e858a29ec38c5e382790ebab8"
        )

        conn = sqlite3.connect("test_listeria_innocua/test_listeria_innocua.db")
        cursor = conn.cursor()

        # Fetch column names and sort them
        cursor.execute(f"PRAGMA table_info(METADATA);")
        columns = [row[1] for row in cursor.fetchall()]
        columns.sort()  # Ensure column order is deterministic

        # Query the data with sorted columns and rows
        sorted_columns = ", ".join(columns)
        cursor.execute(
            f"SELECT {sorted_columns} FROM METADATA ORDER BY {sorted_columns}"
        )
        rows = cursor.fetchall()

        # Compute hash
        hasher = hashlib.sha256()
        for row in rows:
            hasher.update(str(row).encode("utf-8"))  # Hash each row
        conn.close()
        actual_sqlite_sha = hasher.hexdigest()
        hasher = hashlib.sha256()
        database_sig = load_file_as_signatures(
            "test_listeria_innocua/test_listeria_innocua.sig"
        )
        database_sig_sorted = sorted(database_sig, key=lambda x: x.name)
        with open("test_listeria_innocua/test_listeria_innocua.sig.sorted", "wb") as f:
            save_signatures(database_sig_sorted, fp=f)
        with open("test_listeria_innocua/test_listeria_innocua.sig.sorted", "rb") as f:
            buf = f.read()
            hasher.update(buf)
        actual_signature_sha = hasher.hexdigest()
        self.assertEqual(actual_sqlite_sha, expected_sqlite_sha)

        self.assertEqual(actual_signature_sha, expected_signature_sha)
        subprocess.run(
            ["datasets", "download", "genome", "accession", "GCA_022617975.1"]
        )
        # unzip the downloaded file
        with zipfile.ZipFile("ncbi_dataset.zip", "r") as zip_ref:
            zip_ref.extractall("ncbi_dataset")
        subprocess.run(
            [
                "mashpit",
                "query",
                "ncbi_dataset/ncbi_dataset/data/GCA_022617975.1/GCA_022617975.1_PDT001269761.1_genomic.fna",
                "test_listeria_innocua",
            ]
        )
        expected_sha = (
            "82688f582fd95b249f8a9511a243c0b86433897c45ee7d714ea3d5610ac399a9"
        )
        hasher = hashlib.sha256()
        # read the output file
        output_file = pd.read_csv("GCA_022617975_output.csv")
        # sort the output file by PDS_acc
        output_file.sort_values(by=["PDS_acc"], inplace=True)
        # drop the first column
        output_file.drop(output_file.columns[0], axis=1, inplace=True)
        output_file.to_csv("GCA_022617975_output.csv.sorted", index=False)
        with open("GCA_022617975_output.csv.sorted", "rb") as f:
            buf = f.read()
            hasher.update(buf)
        actual_sha = hasher.hexdigest()
        self.assertEqual(actual_sha, expected_sha)


class TestBuildAccession(unittest.TestCase):
    def setUp(self):
        return

    def tearDown(self):
        shutil.rmtree("test_accession")
        os.remove("test_accession_list")
        for file in os.listdir():
            if file.endswith(".log"):
                os.remove(file)

    def test_build_accession(self):
        # generate the test accession file
        accession_list = [
            "SAMN20822594",
        ]
        with open("test_accession_list", "w") as f:
            for accession in accession_list:
                f.write(accession + "\n")
        subprocess.run(
            [
                "mashpit",
                "build",
                "accession",
                "test_accession",
                "--list",
                "test_accession_list",
                "--email",
                "test@example.com",
            ]
        )

        expected_sqlite_sha = (
            "627e3cd4415e559bacde537c9ce60ee639c01f28c8a0bfca7365bdfa5573c8b0"
        )
        expected_signature_sha = (
            "2cdf077e256ada00a4d13e1cf5a22be945b713770da81b1de804ef2cc524b10b"
        )
        conn = sqlite3.connect("test_accession/test_accession.db")
        cursor = conn.cursor()

        # Fetch column names and sort them
        cursor.execute(f"PRAGMA table_info(METADATA);")
        columns = [row[1] for row in cursor.fetchall()]
        columns.sort()  # Ensure column order is deterministic

        # Query the data with sorted columns and rows
        sorted_columns = ", ".join(columns)
        cursor.execute(
            f"SELECT {sorted_columns} FROM METADATA ORDER BY {sorted_columns}"
        )
        rows = cursor.fetchall()

        # Compute hash
        hasher = hashlib.sha256()
        for row in rows:
            hasher.update(str(row).encode("utf-8"))  # Hash each row
        conn.close()
        actual_sqlite_sha = hasher.hexdigest()
        hasher = hashlib.sha256()
        with open("test_accession/test_accession.sig", "rb") as f:
            buf = f.read()
            hasher.update(buf)
        actual_signature_sha = hasher.hexdigest()
        self.assertEqual(actual_sqlite_sha, expected_sqlite_sha)
        self.assertEqual(actual_signature_sha, expected_signature_sha)


class TestCompareTables(unittest.TestCase):

    def test_compare_tables(self):
        # Define input data
        database_cluster_center = pd.DataFrame(
            {
                "PDS_acc_no_suffix": ["PDS_001", "PDS_002", "PDS_003"],
                "PDS_acc": ["acc1", "acc2.1", "acc3"],
                "asm_acc": ["asm1", "asm2", "asm3"],
            }
        )
        latest_cluster_center_metadata = pd.DataFrame(
            {
                "PDS_acc_no_suffix": ["PDS_001", "PDS_002", "PDS_004"],
                "PDS_acc": ["acc1", "acc2.2", "acc4"],
                "asm_acc": ["asm1", "asm2", "asm4"],
            }
        )
        latest_cluster_center = pd.DataFrame(
            {
                "PDS_acc_no_suffix": ["PDS_001", "PDS_002", "PDS_004"],
                "PDS_acc": ["acc1", "acc2.2", "acc4"],
                "asm_acc": ["asm1", "asm2", "asm4"],
            }
        )

        # Call the function to compare the tables
        pds_to_remove, asm_acc_remove, pds_to_add, asm_acc_add = compare_tables(
            database_cluster_center, latest_cluster_center_metadata
        )

        # Assert that the expected changes were identified
        self.assertEqual(pds_to_remove, ["PDS_003", "PDS_002"])
        self.assertEqual(asm_acc_remove, ["asm3", "asm2"])
        self.assertEqual(pds_to_add, ["PDS_002", "PDS_004"])
        self.assertEqual(asm_acc_add, ["asm2", "asm4"])


class TestMashpitWebserver(unittest.TestCase):
    def test_webserver_runs(self):
        """Test if the `mashpit webserver` command runs without issues."""
        # Start the webserver as a subprocess
        process = subprocess.Popen(
            ["mashpit", "webserver"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        try:
            # Wait for a few seconds to see if it crashes
            return_code = process.poll()
            if return_code is not None:
                # If the process exits early, capture stderr
                stdout, stderr = process.communicate()
                self.fail(
                    f"`mashpit webserver` exited early with code {return_code}:\n"
                    f"STDOUT: {stdout.decode()}\n"
                    f"STDERR: {stderr.decode()}"
                )
        finally:
            # Terminate the subprocess after the test
            os.kill(process.pid, signal.SIGTERM)
            process.wait()


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
import sys
import argparse

from mashpit import __version__
from mashpit import build
from mashpit import update
from mashpit import query
from mashpit import webserver


def commandToArgs(commandline):
    parser = argparse.ArgumentParser(
        description="A sketch-based surveillance platform."
    )
    parser.add_argument(
        "-v", "--version", action="version", version="%(prog)s " + __version__
    )

    # subparsers for mashpit
    subparsers = parser.add_subparsers(metavar="command")
    subparsers.required = True
    subparser_build = subparsers.add_parser("build", help="Build mashpit database")
    subparser_update = subparsers.add_parser("update", help="Update a database")
    subparser_query = subparsers.add_parser(
        "query", help="Query for the most similar isolates"
    )
    subparser_webserver = subparsers.add_parser(
        "webserver", help="Start a local web server"
    )

    # arguments for mashpit build
    subparser_build.add_argument(
        "type", type=str, help="mashpit database type", choices=["taxon", "accession"]
    )
    subparser_build.add_argument("name", type=str, help="mashpit database name")
    subparser_build.add_argument("--quiet", action="store_true", help="disable logs")
    subparser_build.add_argument(
        "--number",
        type=int,
        default=1000,
        help="maximum number of hashes for sourmash, default is 1000",
    )
    subparser_build.add_argument(
        "--ksize", type=int, default=31, help="kmer size for sourmash, default is 31"
    )
    subparser_build.add_argument("--species", type=str, help="species name")
    subparser_build.add_argument("--email", type=str, help="Entrez email")
    subparser_build.add_argument("--key", type=str, help="Entrez api key")
    subparser_build.add_argument(
        "--pd_version",
        type=str,
        help="a specified Pathogen Detection version (PDG accession). Default is the latest.",
    )
    subparser_build.add_argument(
        "--list", type=str, help="Path to a list of NCBI BioSample accessions"
    )
    subparser_build.set_defaults(func=build.build)

    # "update" subcommand
    # TODO: remove the name argument? because the name is already in the database folder
    subparser_update.add_argument(
        dest="database", type=str, help="path for the database folder"
    )
    subparser_update.add_argument(dest="name", type=str, help="database name")
    subparser_update.add_argument(
        "--metadata", type=str, help="metadata file in csv format"
    )
    subparser_update.add_argument("--quiet", action="store_true", help="disable logs")
    subparser_update.set_defaults(func=update.update)

    # "query" subcommand
    subparser_query.add_argument(
        dest="sample", type=str, help="file path to the query sample"
    )
    subparser_query.add_argument(
        dest="database", type=str, help="path to the database folder"
    )
    subparser_query.add_argument(
        "--number",
        type=int,
        help="number of isolates in the query output, default is 200",
        default=200,
    )
    subparser_query.add_argument(
        "--threshold",
        type=float,
        help="minimum jaccard similarity for mashtree, default is 0.85",
        default=0.85,
    )
    subparser_query.add_argument(
        "--annotation", type=str, help="mashtree tip annotation, default is none"
    )
    subparser_query.set_defaults(func=query.query)

    # "webserver" subcommand
    subparser_webserver.set_defaults(func=webserver.webserver)

    args = parser.parse_args(commandline)
    return args


def main():
    if len(sys.argv[1:]) < 1:
        print(
            "Subcommand is required to run mashpit. Use -h or --help to show help information.\n"
        )
        print("Subcommand options:\n")

        exit(0)
    args = commandToArgs(sys.argv[1:])
    args.func(args)
    return



#!/usr/bin/env python3
import os
import time
import glob
import csv
import itertools
import zipfile
import screed
import shutil
import logging
import sqlite3
import requests
import psutil
import subprocess

import urllib.request
import dask.dataframe as dd
import pandas as pd
import xml.etree.ElementTree as ET

from tqdm import tqdm
from Bio import Entrez
from html.parser import HTMLParser
from sourmash import SourmashSignature, save_signatures, load_one_signature, MinHash


def log_memory_usage(stage):
    process = psutil.Process()
    mem_info = process.memory_info()
    logging.info(
        f"{stage} - RSS: {mem_info.rss / 1024 ** 2:.2f} MB, VMS: {mem_info.vms / 1024 ** 2:.2f} MB"
    )


# create connection to sqlite db file
def create_connection(sql_path):
    conn = sqlite3.connect(sql_path)
    return conn


# create sqlite3 database with the tables
def create_database(conn):
    # metadata and accessions of a sample
    metadata_table = """CREATE TABLE IF NOT EXISTS METADATA (
                              biosample_acc    TEXT PRIMARY KEY, 
                              taxid            INTEGER,
                              strain           TEXT, 
                              collected_by     TEXT,
                              collection_date  TEXT,
                              geo_loc_name     TEXT,
                              isolation_source TEXT,
                              lat_lon          TEXT,
                              serovar          TEXT,
                              sub_species      TEXT,
                              species          TEXT,
                              genus            TEXT,
                              host             TEXT,
                              host_disease     TEXT,
                              outbreak         TEXT,
                              srr              TEXT,
                              PDT_acc          TEXT,
                              PDS_acc          TEXT,
                              asm_acc          TEXT
                        );"""
    # a description table to keep a record of the dabtabase type, (species and PD version if it is a standard database)
    description_table = """CREATE TABLE IF NOT EXISTS DESC (
                              name    TEXT PRIMARY KEY, 
                              value            TEXT
                        );"""
    # create tables
    c = conn.cursor()
    c.execute(metadata_table)
    c.execute(description_table)
    conn.commit()


class PDGParser(HTMLParser):
    list = []

    def handle_data(self, data):
        if str(data).endswith("tsv"):
            self.list.append(data)


class Species_Name_Parser(HTMLParser):
    list = []

    def handle_data(self, data):
        if str(data).endswith("/"):
            self.list.append(data.strip("/"))


class DownloadProgressBar(tqdm):
    def update_to(self, b=1, bsize=1, tsize=None):
        if tsize is not None:
            self.total = tsize
        self.update(b * bsize - self.n)


def download_url(url, output_path):
    with DownloadProgressBar(
        unit="B",
        unit_scale=True,
        miniters=1,
        unit_divisor=1024,
        desc=url.split("/")[-1],
        leave=False,
    ) as t:
        urllib.request.urlretrieve(url, filename=output_path, reporthook=t.update_to)


def download_metadata(pathogen_name, pd_version, tmp_folder):
    logging.info("Downloading metadata files...")
    time_start_download = time.time()
    # get the metadata and SNP distance files
    if pd_version is not None:
        r = requests.get(
            f"https://ftp.ncbi.nlm.nih.gov/pathogen/Results/{pathogen_name}/{pd_version}/Metadata/"
        )
        if r.status_code == 404:
            logging.error("invalid PDG accession")
            exit(0)
        elif r.status_code == 200:
            url_metadata_folder = f"https://ftp.ncbi.nlm.nih.gov/pathogen/Results/{pathogen_name}/{pd_version}/Metadata/"
    else:
        url_metadata_folder = f"https://ftp.ncbi.nlm.nih.gov/pathogen/Results/{pathogen_name}/latest_snps/Metadata/"
    response = urllib.request.urlopen(url_metadata_folder)
    metadata_html = response.read().decode("utf-8")
    parser = PDGParser()
    parser.feed(metadata_html)
    metadata_file_name = parser.list[0]
    pdg_acc = metadata_file_name.strip(".metadata.tsv")
    distance_file = f"{pdg_acc}.reference_target.SNP_distances.tsv"
    isolate_pds_file = f"{pdg_acc}.reference_target.all_isolates.tsv"
    parser.close()
    url_metadata = url_metadata_folder + metadata_file_name
    if pd_version is not None:
        url_snp_distances = f"https://ftp.ncbi.nlm.nih.gov/pathogen/Results/{pathogen_name}/{pd_version}/Clusters/{distance_file}"
        url_isolate_pds = f"https://ftp.ncbi.nlm.nih.gov/pathogen/Results/{pathogen_name}/{pd_version}/Clusters/{isolate_pds_file}"
    else:
        url_snp_distances = f"https://ftp.ncbi.nlm.nih.gov/pathogen/Results/{pathogen_name}/latest_snps/Clusters/{distance_file}"
        url_isolate_pds = f"https://ftp.ncbi.nlm.nih.gov/pathogen/Results/{pathogen_name}/latest_snps/Clusters/{isolate_pds_file}"
    # check if the metadata file exists
    if not os.path.isfile(os.path.join(tmp_folder, metadata_file_name)):
        logging.info(f"Downloading metadata file from {url_metadata_folder}")
        download_url(url_metadata, os.path.join(tmp_folder, metadata_file_name))
    else:
        logging.info(f"Using existing metadata file {metadata_file_name}")
    # check if the SNP distance file exists
    if not os.path.isfile(os.path.join(tmp_folder, distance_file)):
        logging.info(f"Downloading pairwise SNP distance file from {url_snp_distances}")
        download_url(url_snp_distances, os.path.join(tmp_folder, distance_file))
    else:
        logging.info(f"Using existing pairwise SNP distance file {distance_file}")
    # check if the isolate PDS acc file exists
    if not os.path.isfile(os.path.join(tmp_folder, isolate_pds_file)):
        logging.info(f"Downloading isolates_PDS_acc file from {url_isolate_pds}")
        download_url(url_isolate_pds, os.path.join(tmp_folder, isolate_pds_file))
    else:
        logging.info(f"Using existing isolates_PDS_acc file {isolate_pds_file}")

    time_end_download = time.time()
    logging.info(
        f"Downloaded metadata and distance files in {round(time_end_download-time_start_download,2)} seconds."
    )
    # return the metadata file name
    return metadata_file_name, isolate_pds_file


def calculate_centroid(df_metadata, pdg_acc, tmp_folder):
    logging.info("Calculating centroid...")
    time_start_calculate_centroid = time.time()
    # drop columns to reduce memory usage
    input_file_name = f"{pdg_acc}.reference_target.SNP_distances.tsv"
    output_file_name = f"{pdg_acc}_distance.tsv"
    input_file = os.path.join(tmp_folder, input_file_name)
    output_file = os.path.join(tmp_folder, output_file_name)
    columns_to_select = [0, 4, 8, 11]

    with open(input_file, "r") as f_in, open(output_file, "w") as f_out:
        reader = csv.reader(f_in, delimiter="\t")
        writer = csv.writer(f_out, delimiter="\t")
        for row in reader:
            selected_cols = [row[i] for i in columns_to_select]
            writer.writerow(selected_cols)

    # filter out isolates without asm_acc
    target_columns = ["target_acc_1", "target_acc_2"]
    distance_filtered_name = f"{pdg_acc}_filtered_distance.tsv"
    distance_filtered = os.path.join(tmp_folder, distance_filtered_name)
    # Use the csv module to open the TSV file and create a reader object
    with open(output_file, "r") as distance_file, open(
        distance_filtered, "w"
    ) as filtered_distance_file:

        reader = csv.reader(distance_file, delimiter="\t")
        writer = csv.writer(filtered_distance_file, delimiter="\t")

        # Extract the header row from the reader object and write it to the new file
        header = next(reader)
        writer.writerow(header)

        # Find the indices of the target columns in the header row
        target_column_indices = [header.index(column) for column in target_columns]

        # Iterate over each row in the reader object and write it to the new file if it contains the target columns
        target_values = set(df_metadata["target_acc"])
        for row in reader:
            if all(row[index] in target_values for index in target_column_indices):
                writer.writerow(row)

    # Define the column to group by
    group_by_column = "PDS_acc"
    # two lists to be added into the database
    cluster_list_db = []
    centroid_list_db = []
    # Open the TSV file using the csv module and read it line by line
    with open(distance_filtered, "r") as file:
        reader = csv.reader(file, delimiter="\t")
        header = next(reader)  # read the header row

        # Define a generator expression that yields each row from the file
        rows = (row for row in reader)

        # Use itertools.groupby() to group the rows by the specified column
        for group_value, group_rows in itertools.groupby(
            rows, lambda x: x[header.index(group_by_column)]
        ):
            # Process the rows for the current group
            cluster_list_db.append(group_value)
            all_rows = list(group_rows)
            col1 = [row[0] for row in all_rows]
            col2 = [row[1] for row in all_rows]
            col3 = [row[3] for row in all_rows]
            df_cluster = pd.DataFrame(
                {
                    "target_acc_1": col1,
                    "target_acc_2": col2,
                    "delta_positions_unambiguous": col3,
                }
            )
            df_cluster = pd.concat(
                [
                    df_cluster,
                    df_cluster.rename(
                        columns={
                            "target_acc_1": "target_acc_2",
                            "target_acc_2": "target_acc_1",
                        }
                    ),
                ]
            )
            df_cluster["delta_positions_unambiguous"] = df_cluster[
                "delta_positions_unambiguous"
            ].astype("float")
            df_dask = dd.from_pandas(
                df_cluster, npartitions=4
            )  # Convert the pandas dataframe to a Dask dataframe with 4 partitions
            centroid_target_acc = (
                df_dask.groupby("target_acc_1")["delta_positions_unambiguous"]
                .sum()
                .compute()
                .idxmin()
            )
            centroid_list_db.append(centroid_target_acc)

    df_cluster_center = pd.DataFrame(
        {"PDS_acc": cluster_list_db, "target_acc": centroid_list_db}
    )
    df_cluster_center.to_csv(
        os.path.join(tmp_folder, f"{pdg_acc}_cluster_center.tsv"), sep="\t", index=False
    )
    time_end_calculate_centroid = time.time()
    logging.info(
        f"Calculated centroid in {round(time_end_calculate_centroid-time_start_calculate_centroid,2)} seconds."
    )
    logging.info(f"Number of clusters: {len(cluster_list_db)}")


def download_and_sketch_assembly(gca_acc_list, hash_number, kmer_size, tmp_folder):
    logging.info("Downloading and sketching assemblies...")
    time_start_download_and_sketch_assembly = time.time()
    gca_acc_file = os.path.join(tmp_folder, "gca_list.txt")
    with open(gca_acc_file, "w") as f:
        for item in gca_acc_list:
            f.write("%s\n" % item)
    subprocess.run(
        [
            "datasets",
            "download",
            "genome",
            "accession",
            "--inputfile",
            gca_acc_file,
            "--filename",
            "assembly.zip",
            "--dehydrated",
        ]
    )
    # move the zip file to the tmp folder
    zip_file = os.path.join(tmp_folder, "assembly.zip")
    shutil.move("assembly.zip", zip_file)
    with zipfile.ZipFile(zip_file, "r") as zip_ref:
        zip_ref.extractall(os.path.join(tmp_folder, "assembly"))
    # try to rehydrate the dataset, if failed, try again after 5 seconds
    while True:
        try:
            subprocess.run(
                [
                    "datasets",
                    "rehydrate",
                    "--directory",
                    os.path.join(tmp_folder, "assembly"),
                ],
                check=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            break
        except subprocess.CalledProcessError:
            time.sleep(5)
            logging.error(f"Failed to rehydrate {tmp_folder}/assembly. Trying again...")
    time_end_download = time.time()
    logging.info(
        f"Downloaded {len(gca_acc_list)} assemblies in {round(time_end_download-time_start_download_and_sketch_assembly,2)} seconds."
    )
    os.mkdir(os.path.join(tmp_folder, "signature"))
    for acc in gca_acc_list:
        # TODO: other file name patterns?
        fna_file_list = glob.glob(
            os.path.join(
                tmp_folder, "assembly", "ncbi_dataset", "data", acc, "*_genomic.fna"
            )
        )
        # skip if there are more than one fna files or no fasta file
        if len(fna_file_list) != 1:
            continue
        mh = MinHash(n=hash_number, ksize=kmer_size)
        with screed.open(fna_file_list[0], "r") as f:
            for record in f:
                mh.add_sequence(record.sequence, True)
        signame = acc
        sig = [SourmashSignature(mh, name=signame)]
        with open(os.path.join(tmp_folder, "signature", acc + ".sig"), "w") as f:
            save_signatures(sig, fp=f)
    time_end_sketch = time.time()
    logging.info(
        f"Sketched {len(gca_acc_list)} assemblies in {round(time_end_sketch-time_end_download,2)} seconds."
    )


def merge_sig(args, db_folder, tmp_folder):
    logging.info("Merging all signature files...")
    time_start_merge_sig = time.time()
    all_sig_path = os.path.join(tmp_folder, "signature", "*.sig")
    sig_path_list = glob.glob(all_sig_path)
    siglist = []
    for sig_path in sig_path_list:
        siglist.append(load_one_signature(sig_path))
    sig_file_name = os.path.join(db_folder, args.name + ".sig")
    with open(sig_file_name, "w") as f:
        save_signatures(siglist, fp=f)
    time_end_merge_sig = time.time()
    logging.info(
        f"Merged all signature files in {round(time_end_merge_sig-time_start_merge_sig,2)} seconds."
    )


# import metadata to database directly from pathogen detection metadata dataframe
def import_metadata(df_metadata, df_cluster_center_metadata, conn):
    info = {
        "biosample_acc": None,
        "taxid": None,
        "strain": None,
        "collected_by": None,
        "collection_date": None,
        "geo_loc_name": None,
        "isolation_source": None,
        "lat_lon": None,
        "serovar": None,
        "sub_species": None,
        "species": None,
        "genus": None,
        "host": None,
        "host_disease": None,
        "outbreak": None,
        "srr": None,
        "PDT_acc": None,
        "PDS_acc": None,
        "asm_acc": None,
    }

    for index, row in df_cluster_center_metadata.iterrows():
        info["biosample_acc"] = row["biosample_acc"]
        info["taxid"] = row["taxid"]
        info["strain"] = row["strain"]
        info["collected_by"] = row["collected_by"]
        info["collection_date"] = row["collection_date"]
        info["geo_loc_name"] = row["geo_loc_name"]
        info["isolation_source"] = row["isolation_source"]
        info["lat_lon"] = row["lat_lon"]
        info["serovar"] = row["serovar"]
        info["host"] = row["host"]
        info["host_disease"] = row["host_disease"]
        info["outbreak"] = ",".join(
            list(
                df_metadata[df_metadata["PDS_acc"] == row["PDS_acc"]]
                .dropna(subset=["outbreak"])
                .outbreak.unique()
            )
        )
        info["srr"] = row["Run"]
        info["PDT_acc"] = row["target_acc"]
        info["PDS_acc"] = row["PDS_acc"]
        info["asm_acc"] = row["asm_acc"]
        # check whether the information is missing
        for key in info:
            if info[key] is None:
                info[key] = "missing"

        info_list = list(info.values())

        with conn:
            insert_metadata(conn, info_list)
    return


def insert_metadata(conn, info):
    sql = """INSERT OR IGNORE INTO METADATA(
             biosample_acc,
             taxid,
             strain,
             collected_by,
             collection_date,
             geo_loc_name,
             isolation_source,
             lat_lon,
             serovar,
             sub_species,
             species,
             genus,
             host,
             host_disease,
             outbreak,
             srr,
             PDT_acc,
             PDS_acc,
             asm_acc) 
             VALUES(:biosample_acc, :taxid, :strain, :collected_by, :collection_date,
                    :geo_loc_name, :isolation_source, :lat_lon, :serovar, :sub_species,
                    :species, :genus, :host, :host_disease, :outbreak, :srr,
                    :PDT_acc, :PDS_acc, :asm_acc);"""
    c = conn.cursor()
    c.execute(sql, info)


def prepare(args):
    # check if datasets is installed
    if not shutil.which("datasets"):
        logging.error(
            "NCBI datasets is not installed. Please install it from https://www.ncbi.nlm.nih.gov/datasets/docs/v2/download-and-install/"
        )
        exit(1)
    t = time.localtime()
    current_time = time.strftime("%Y%m%d%H%M%S", t)
    print_handler = logging.StreamHandler()
    file_handler = logging.FileHandler(f"mashpit-{current_time}.log")
    print_handler.setLevel(logging.INFO)
    file_handler.setLevel(logging.DEBUG)
    if args.quiet:
        print_handler.setLevel(logging.ERROR)
    logging.basicConfig(
        format="%(asctime)s [%(levelname)s] %(message)s",
        level=logging.DEBUG,
        datefmt="%d-%b-%y %H:%M:%S",
        handlers=[print_handler, file_handler],
    )

    # print out database infomation
    logging.info("-" * 50)
    if args.type == "taxon":
        logging.info("Database type: Taxon")
        if args.species is None:
            logging.error("Taxon name not provided")
            exit(1)
        logging.info(f"Species: {args.species}")
    if args.type == "accession":
        logging.info("Database type: Accession")
        if args.list is None:
            logging.error("Accession list not provided")
            exit(1)
        logging.info(f"Biosample accession list: {args.list}")
        if args.email == None:
            logging.error("Entrez email not provided.")
            exit(1)
    if args.name is None:
        logging.error("Database name not provided")
        exit(1)
    logging.info(f"Database name: {args.name}")
    logging.info(f"Maximum number of hashes: {args.number}")
    logging.info(f"Kmer size: {args.ksize}")
    logging.info("-" * 50)

    # create a database folder
    cwd = os.getcwd()
    db_folder = os.path.join(cwd, args.name)
    if os.path.isdir(db_folder):
        logging.error(f"Folder {db_folder} already exists.")
        exit(1)
    os.mkdir(db_folder)
    logging.info(f"Database folder created at {db_folder}")

    # create a tmp folder
    tmp_folder = os.path.join(cwd, "tmp")
    if os.path.isdir(tmp_folder):
        shutil.rmtree(tmp_folder)
    os.mkdir(tmp_folder)

    # create a sqlite database
    sql_path = os.path.join(db_folder, f"{args.name}.db")
    conn = create_connection(sql_path)
    create_database(conn)

    return db_folder, tmp_folder, conn


def build_taxon(args):
    # prepare the database folder and sqlite database
    db_folder, tmp_folder, conn = prepare(args)
    log_memory_usage("start")
    # format the pathogen name
    pathogen_name = args.species.strip().replace(" ", "_")
    # validate the species name
    results_response = urllib.request.urlopen(
        "https://ftp.ncbi.nlm.nih.gov/pathogen/Results/"
    )
    results_html = results_response.read().decode("utf-8")
    results_parser = Species_Name_Parser()
    results_parser.feed(results_html)
    pathogen_name_lower = pathogen_name.lower()
    for item in results_parser.list:
        # convert the list item to lowercase
        item_lower = item.lower()
        # compare the lowercase versions of the string and the list item
        if pathogen_name_lower == item_lower:
            pathogen_name = item
            break
    else:
        logging.error(
            f"Taxon {pathogen_name} not available on Pathogen Detection. Check for available names at https://ftp.ncbi.nlm.nih.gov/pathogen/Results/"
        )
        exit(1)
    logging.info(f"Taxon name validated. Using {pathogen_name} as the taxon name.")
    # download the latest PD metadata files and get the PDG accession
    metadata_file_name, isolate_pds_file = download_metadata(
        pathogen_name, args.pd_version, tmp_folder
    )
    log_memory_usage("After downloading files")
    df_metadata = pd.read_csv(
        os.path.join(tmp_folder, metadata_file_name), header=0, sep="\t"
    )
    df_isolate_pds = pd.read_csv(
        os.path.join(tmp_folder, isolate_pds_file), header=0, sep="\t"
    )
    # add cluster accession to metadata
    df_metadata = df_metadata.join(
        df_isolate_pds[["target_acc", "PDS_acc"]].set_index("target_acc"),
        on="target_acc",
    )
    pdg_acc = metadata_file_name.replace(".metadata.tsv", "")
    # filter out isolates without genome assembly
    df_metadata_asm = df_metadata[~df_metadata["asm_acc"].isnull()]
    # calculate the centroid for each cluster
    calculate_centroid(df_metadata_asm, pdg_acc, tmp_folder)
    df_cluster_center = pd.read_csv(
        os.path.join(tmp_folder, f"{pdg_acc}_cluster_center.tsv"), sep="\t"
    )
    df_cluster_center_metadata = df_cluster_center.join(
        df_metadata_asm.drop("PDS_acc", axis=1).set_index("target_acc"), on="target_acc"
    )
    log_memory_usage("After centroid calculation")
    # download the centroid assembly using NCBI datasets
    gca_acc_list = df_cluster_center_metadata["asm_acc"].to_list()
    hash_number = args.number
    kmer_size = args.ksize
    download_and_sketch_assembly(gca_acc_list, hash_number, kmer_size, tmp_folder)
    log_memory_usage("After downloading assemblies")
    # merge all signature files
    merge_sig(args, db_folder, tmp_folder)
    log_memory_usage("After merging signature files")
    # import metadata to database
    import_metadata(df_metadata, df_cluster_center_metadata, conn)
    log_memory_usage("After importing metadata")
    # add description to database
    c = conn.cursor()
    c.execute("INSERT INTO DESC VALUES ('Type','Taxonomy');")
    c.execute(f"INSERT INTO DESC VALUES ('Species','{pathogen_name}');")
    c.execute(f"INSERT INTO DESC VALUES ('Version','{pdg_acc}');")
    c.execute(f"INSERT INTO DESC VALUES ('Hash_number','{str(hash_number)}');")
    c.execute(f"INSERT INTO DESC VALUES ('Kmer_size','{str(kmer_size)}');")
    conn.commit()

    # delete temp files and folders
    logging.info("Deleting temp file and folders")
    shutil.rmtree(tmp_folder)
    logging.info("Complete.")


def build_accession(args):
    db_folder, tmp_folder, conn = prepare(args)
    Entrez.email = args.email
    if args.key != None:
        Entrez.api_key = args.key
    # check if the list file exists
    if not os.path.isfile(args.list):
        logging.error(f"File {args.list} not found.")
        exit(1)
    # download assembly and get metadata
    gca_acc_list = []
    f = open(args.list, "r")
    biosample_list = f.readlines()
    for biosample in biosample_list:
        # if api key is not provided, wait for 1 second
        if args.key == None:
            time.sleep(3)
        # check if the assembly is available on NCBI
        esearch_handle = Entrez.esearch(db="assembly", term=biosample, retmax="3")
        esearch_record = Entrez.read(esearch_handle)
        # skip if no assembly found
        if len(esearch_record["IdList"]) == 0:
            continue
        # get assembly accession
        if args.key == None:
            time.sleep(3)
        esummary_handle = Entrez.esummary(
            db="assembly", id=esearch_record["IdList"][0], report="full"
        )
        esummary_record = Entrez.read(esummary_handle)
        asm_acc = esummary_record["DocumentSummarySet"]["DocumentSummary"][0][
            "AssemblyAccession"
        ]
        gca_acc_list.append(asm_acc)
        # get metadata
        # if api key is not provided, wait for 1 second
        if args.key == None:
            time.sleep(3)
        esearch_handle = Entrez.esearch(db="biosample", term=biosample, retmax="3")
        esearch_record = Entrez.read(esearch_handle)
        if args.key == None:
            time.sleep(3)
        efetch_handle = Entrez.efetch(db="biosample", id=esearch_record["IdList"][0])
        efetch_record = efetch_handle.read()
        root = ET.fromstring(efetch_record)
        info = {
            "biosample_acc": None,
            "taxid": None,
            "strain": None,
            "collected_by": None,
            "collection_date": None,
            "geo_loc_name": None,
            "isolation_source": None,
            "lat_lon": None,
            "serovar": None,
            "sub_species": None,
            "species": None,
            "genus": None,
            "host": None,
            "host_disease": None,
            "outbreak": None,
            "srr": None,
            "PDT_acc": None,
            "PDS_acc": None,
            "asm_acc": None,
        }
        info["biosample_acc"] = biosample
        info["asm_acc"] = asm_acc
        for domain in root[0]:
            if domain.tag == "Attributes":
                for sub_attrib in domain:
                    if sub_attrib.attrib["attribute_name"] in info:
                        info[sub_attrib.attrib["attribute_name"]] = sub_attrib.text
        insert_metadata(conn, info)
    f.close()
    hash_number = args.number
    kmer_size = args.ksize
    download_and_sketch_assembly(gca_acc_list, hash_number, kmer_size, tmp_folder)
    # merge all generated signature files
    merge_sig(args, db_folder, tmp_folder)
    c = conn.cursor()
    c.execute("INSERT INTO DESC VALUES ('Type','Accession');")
    c.execute("INSERT INTO DESC VALUES ('Hash_number','" + str(hash_number) + "');")
    c.execute("INSERT INTO DESC VALUES ('Kmer_size','" + str(kmer_size) + "');")
    conn.commit()
    logging.info("Deleting tmp folders")
    shutil.rmtree(tmp_folder)
    logging.info("Complete.")


def build(args):
    if args.type == "taxon":
        build_taxon(args)
    elif args.type == "accession":
        build_accession(args)



#!/usr/bin/env python3
import os
import glob
import logging
import ntpath
import screed
import sourmash
import heapq
import time
import pandas as pd

from skbio import DistanceMatrix
from skbio.tree import nj
from phytreeviz import TreeViz
from mashpit.build import create_connection
from operator import itemgetter
from sourmash import (
    SourmashSignature,
    save_signatures,
    load_one_signature,
    load_file_as_signatures,
)


def get_query_sig(query_path, query_name, hash_number, kmer_size):
    genome = query_path
    mh = sourmash.MinHash(n=hash_number, ksize=kmer_size)
    for record in screed.open(genome):
        mh.add_sequence(record.sequence, True)
    sig = SourmashSignature(mh, name=genome)
    query_sig_path = os.path.join(os.path.dirname(query_path), f"{query_name}.sig")
    with open(query_sig_path, "wt") as fp:
        save_signatures([sig], fp)


def generate_query_table(conn, sorted_asm_similarity_dict):
    c = conn.cursor()
    c.execute("SELECT * FROM METADATA")
    names = [description[0] for description in c.description]

    # List to collect DataFrames
    df_list = []

    for i in sorted_asm_similarity_dict:
        sql_query = pd.read_sql_query(
            f"SELECT * FROM METADATA WHERE asm_acc = '{i}'", conn
        )
        df_query = pd.DataFrame(sql_query, columns=names)
        df_query["similarity_score"] = sorted_asm_similarity_dict[i]
        df_list.append(df_query)

    # Concatenate all DataFrames at once
    output_df = pd.concat(df_list, ignore_index=True)

    # Additional processing (if it is a taxon database, add the link of the SNP cluster to the output)
    c.execute("SELECT value FROM DESC WHERE name = 'Type';")
    db_type = c.fetchone()[0]
    if db_type == "Taxonomy":
        pds_list = output_df["PDS_acc"].to_list()
        cluster_link = [
            f"https://www.ncbi.nlm.nih.gov/pathogens/isolates/#{pds}"
            for pds in pds_list
        ]
        output_df["SNP_tree_link"] = cluster_link

    return output_df


def generate_mashtree(
    output_df, min_similarity, query_name, sig_path, added_annotation, database_sig
):
    # check if the top query similarity is smaller than the threshold
    if float(output_df["similarity_score"].iloc[0]) < min_similarity:
        logging.error(
            "Top query similarity is smaller than the threshold. Mashtree can not be generated. "
        )
        exit(1)
    # select top results that are above the threshold
    acc_list = output_df[output_df["similarity_score"] >= min_similarity][
        "asm_acc"
    ].to_list()
    # if the number of top results is smaller than 2, mashtree can not be generated
    if len(acc_list) < 2:
        logging.error(
            "Number of top results is smaller than 2. Mashtree can not be generated. "
        )
        exit(1)
    leaves = [query_name] + acc_list
    # select database signatures that are present in acc_list
    sigs = []
    for sig in database_sig:
        if sig.name in acc_list:
            sigs.append(sig)
    # sort sigs based on the order of acc_list
    sigs = sorted(sigs, key=lambda x: acc_list.index(x.name))
    matrix = []
    # first row
    first_row = [0]
    for similarity_score in output_df[output_df["similarity_score"] >= 0.85][
        "similarity_score"
    ].to_list():
        first_row.append(1 - similarity_score)
    matrix.append(first_row)
    for acc in acc_list:
        distance_list = []
        distance_list.append(
            1 - (output_df[output_df["asm_acc"] == acc]["similarity_score"].iloc[0])
        )
        for sig in sigs:
            if sig.name == acc:
                query_sig = sig
        for sig in sigs:
            distance_list.append(1 - (query_sig.jaccard(sig)))
        matrix.append(distance_list)

    dm = DistanceMatrix(matrix, leaves)
    newick_str = nj(dm, result_constructor=str)
    with open(f"{query_name}_tree.newick", "w") as f:
        f.write(newick_str)

    # add annotation
    if added_annotation is not None:
        annotated_leaves = []
        for leaf in leaves:
            if leaf in acc_list:
                added_annotation = str(
                    output_df[output_df["asm_acc"] == leaf][added_annotation].iloc[0]
                )
                leaf = leaf + " " + added_annotation
            annotated_leaves.append(leaf)
        dm = DistanceMatrix(matrix, annotated_leaves)
        newick_str = nj(dm, result_constructor=str)
        with open(f"{query_name}_tree.newick", "w") as f:
            f.write(newick_str)
    tv = TreeViz(f"{query_name}_tree.newick")
    tv.set_node_label_props(query_name, color="red")
    tv.savefig(f"{query_name}_tree.png", dpi=300)


def query(args):
    t = time.localtime()
    current_time = time.strftime("%Y%m%d%H%M%S", t)
    print_handler = logging.StreamHandler()
    file_handler = logging.FileHandler(f"mashpit-{current_time}.log")
    print_handler.setLevel(logging.INFO)
    file_handler.setLevel(logging.DEBUG)
    logging.basicConfig(
        format="%(asctime)s [%(levelname)s] %(message)s",
        level=logging.DEBUG,
        datefmt="%d-%b-%y %H:%M:%S",
        handlers=[print_handler, file_handler],
    )
    logging.info("Start querying")
    time_start = time.time()
    number_results = args.number
    min_similarity = args.threshold
    added_annotation = args.annotation
    query_path = os.path.abspath(args.sample)
    query_name = ntpath.basename(query_path)

    # remove the extension if any
    if "." in query_name:
        query_name = query_name.split(".")[0]
    db_folder = os.path.abspath(args.database)
    if not os.path.exists(db_folder):
        logging.error("Database path not found.")
        exit(1)
    # find files ending with .db
    sql_path = glob.glob(os.path.join(db_folder, "*.db"))
    if len(sql_path) == 0:
        logging.error("Database path not found.")
        exit(1)
    if len(sql_path) > 1:
        logging.error("Multiple database files found.")
        exit(1)
    sql_path = sql_path[0]
    # find files ending with .sig
    sig_path = glob.glob(os.path.join(db_folder, "*.sig"))
    if len(sig_path) == 0:
        logging.error("Signature file not found.")
        exit(1)
    if len(sig_path) > 1:
        logging.error("Multiple signature files found.")
        exit(1)
    sig_path = sig_path[0]
    # check if the basename of .db and .sig are the same
    if (
        os.path.basename(sql_path).split(".")[0]
        != os.path.basename(sig_path).split(".")[0]
    ):
        logging.error("Database sql and signature files do not match.")
        exit(1)

    # check the hash number and kmer size in the database
    conn = create_connection(sql_path)
    c = conn.cursor()
    c.execute("SELECT value FROM DESC where name = 'Hash_number';")
    hash_number = int(c.fetchone()[0])
    c.execute("SELECT value FROM DESC where name = 'Kmer_size';")
    kmer_size = int(c.fetchone()[0])

    # sketch the query sample and load the signature
    get_query_sig(query_path, query_name, hash_number, kmer_size)
    query_sig_path = os.path.join(os.path.dirname(query_path), f"{query_name}.sig")
    query_sig = load_one_signature(query_sig_path)

    time_finish_sketch = time.time()
    logging.info(
        f"Query sample sketched in {time_finish_sketch-time_start:.2f} seconds"
    )

    asm_similarity_dict = {}
    database_sig = list(load_file_as_signatures(sig_path))

    time_load_database_sig = time.time()
    logging.info(
        f"Database signatures loaded in {time_load_database_sig-time_finish_sketch:.2f} seconds"
    )

    for sig in database_sig:
        similarity = query_sig.jaccard(sig)
        asm_similarity_dict[str(sig)] = similarity

    time_calculate_similarity = time.time()
    logging.info(
        f"Jaccard similarity calculated in {time_calculate_similarity-time_load_database_sig:.2f} seconds"
    )

    # get the top results
    top_items = heapq.nlargest(
        number_results, asm_similarity_dict.items(), key=itemgetter(1)
    )
    sorted_asm_similarity_dict = dict(top_items)

    time_sort = time.time()
    logging.info(
        f"Top {number_results} results sorted in {time_sort-time_calculate_similarity:.2f} seconds"
    )

    output_df = generate_query_table(conn, sorted_asm_similarity_dict)
    output_df.to_csv(query_name + "_output.csv", index=True)
    generate_mashtree(
        output_df, min_similarity, query_name, sig_path, added_annotation, database_sig
    )
    time_mashtree = time.time()
    logging.info(f"Mashtree generated in {time_mashtree-time_sort:.2f} seconds")



import os
import threading
import subprocess
import shutil
import pandas as pd
from flask import Flask, render_template, request, redirect, url_for

app = Flask(
    __name__,
    template_folder=os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "templates"
    ),
)

# Lock for thread-safe operations on global variables
lock = threading.Lock()

# Global variables to store the output of the classification and status
mashpit_out = []
mashpit_status = "idle"


def run_mashpit(asm_file_path):
    global mashpit_out
    global mashpit_status
    try:
        with lock:
            mashpit_status = "in progress"
        upload_folder = os.path.join(app.config["UPLOAD_FOLDER"])
        subprocess.run(["mashpit", "query", asm_file_path, upload_folder], check=True)
        query_name = os.path.basename(asm_file_path).split(".")[0]
        # move the output to the static folder
        static_folder = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "static"
        )
        # check if file exists and remove it
        if os.path.exists(os.path.join(static_folder, f"{query_name}_output.csv")):
            os.remove(os.path.join(static_folder, f"{query_name}_output.csv"))
        if os.path.exists(os.path.join(static_folder, f"{query_name}_tree.png")):
            os.remove(os.path.join(static_folder, f"{query_name}_tree.png"))
        shutil.move(f"{query_name}_output.csv", static_folder)
        shutil.move(f"{query_name}_tree.png", static_folder)

        df_output = pd.read_csv(os.path.join(static_folder, f"{query_name}_output.csv"))
        # remove first column
        df_output = df_output.iloc[:, 1:]
        # put column 'asm_acc', 'similarity_score' and 'SNP_tree_link' to the front
        df_output = df_output[
            ["asm_acc", "similarity_score", "SNP_tree_link"]
            + [
                col
                for col in df_output.columns
                if col not in ["asm_acc", "similarity_score", "SNP_tree_link"]
            ]
        ]

        image_path = f"{query_name}_tree.png"
        df_output_html = df_output.to_html(
            index=False, classes="table table-striped table-hover"
        )

        with lock:
            mashpit_out = [image_path, df_output_html]
            mashpit_status = "complete"

    except Exception as e:
        with lock:
            mashpit_status = f"error: {str(e)}"


@app.route("/", methods=["GET"])
def upload():
    return render_template("upload.html")


@app.route("/submit", methods=["POST"])
def submit():
    uploaded_files = request.files
    paths = []
    for key in ["formFile_db", "formFile_sig", "formFile_assembly"]:
        file = uploaded_files[key]
        file_path = os.path.join(app.config["UPLOAD_FOLDER"], file.filename)
        file.save(file_path)
        paths.append(file_path)

    global mashpit_thread
    global mashpit_status

    with lock:
        mashpit_status = "preparing"
        mashpit_thread = threading.Thread(target=run_mashpit, args=(paths[2],))
        mashpit_thread.start()

    return redirect(url_for("loading"))


@app.route("/loading")
def loading():
    global mashpit_status
    with lock:
        current_status = mashpit_status
    return render_template("loading.html", status=current_status)


@app.route("/result")
def result():
    global mashpit_out
    with lock:
        image_path, df_output_html = mashpit_out
    return render_template(
        "result.html", image_path=image_path, df_output_html=df_output_html
    )


@app.route("/status")
def status():
    global mashpit_status
    with lock:
        current_status = mashpit_status
    return current_status


def webserver(args):
    app.config["UPLOAD_FOLDER"] = "uploads"
    # if the folder exists, clear it
    if os.path.exists(app.config["UPLOAD_FOLDER"]):
        for file in os.listdir(app.config["UPLOAD_FOLDER"]):
            os.remove(os.path.join(app.config["UPLOAD_FOLDER"], file))
    else:
        os.makedirs(app.config["UPLOAD_FOLDER"], exist_ok=True)
    app.run(debug=True, host="0.0.0.0", port=8080)



import os
import time
import pandas as pd
import glob
import urllib.request
import logging
import shutil

from sourmash import load_file_as_signatures, save_signatures, load_one_signature
from mashpit.build import (
    create_connection,
    PDGParser,
    download_metadata,
    calculate_centroid,
    import_metadata,
    download_and_sketch_assembly,
)


def check_database_version(db_pathogen_name, db_version):
    # check if there is a new version
    url_metadata_folder = (
        "https://ftp.ncbi.nlm.nih.gov/pathogen/Results/"
        + db_pathogen_name
        + "/latest_snps/Metadata/"
    )
    response = urllib.request.urlopen(url_metadata_folder)
    metadata_html = response.read().decode("utf-8")
    parser = PDGParser()
    parser.feed(metadata_html)
    metadata_file_name = parser.list[0]
    latest_pdg_acc = metadata_file_name.strip(".metadata.tsv")
    if db_version == latest_pdg_acc:
        logging.error("%s is already the latest version" % db_version)
        exit(0)
    else:
        logging.info(
            "Updating %s from %s to %s" % (db_pathogen_name, db_version, latest_pdg_acc)
        )

    return latest_pdg_acc


def compare_tables(database_cluster_center, latest_cluster_center_metadata):
    logging.info("Comparing two databases")
    pds_to_remove = []
    asm_acc_remove = []
    pds_to_add = []
    asm_acc_add = []
    # 1. If PDS is no longer in latest metadata, remove old metadata record and signature
    pds_only_in_old = list(
        set(database_cluster_center["PDS_acc_no_suffix"].to_list())
        - set(latest_cluster_center_metadata["PDS_acc_no_suffix"].to_list())
    )
    if len(pds_only_in_old) > 0:
        for pds in pds_only_in_old:
            pds_to_remove.append(pds)
        for pds in database_cluster_center[
            database_cluster_center["PDS_acc_no_suffix"].isin(pds_to_remove)
        ]["asm_acc"].to_list():
            asm_acc_remove.append(pds)
    # 2. If PDS is present in both tables, check if the centroid that has changed
    pds_common = list(
        set(database_cluster_center["PDS_acc_no_suffix"].to_list()).intersection(
            latest_cluster_center_metadata["PDS_acc_no_suffix"].to_list()
        )
    )
    if len(pds_common) > 0:
        pds_changed = []
        for pds in pds_common:
            database_center = database_cluster_center[
                database_cluster_center["PDS_acc_no_suffix"] == pds
            ]["PDS_acc"].iloc[0]
            latest_center = latest_cluster_center_metadata[
                latest_cluster_center_metadata["PDS_acc_no_suffix"] == pds
            ]["PDS_acc"].iloc[0]
            if database_center != latest_center:
                pds_changed.append(pds)
        if len(pds_changed) > 0:
            for pds in pds_changed:
                pds_to_add.append(pds)
                pds_to_remove.append(pds)
            for pds in database_cluster_center[
                database_cluster_center["PDS_acc_no_suffix"].isin(pds_changed)
            ]["asm_acc"].to_list():
                asm_acc_remove.append(pds)
                asm_acc_add.append(pds)
    # 3. If PDS is only in the new table, add metadata and download signature files
    pds_new = list(
        set(latest_cluster_center_metadata["PDS_acc_no_suffix"].to_list())
        - set(database_cluster_center["PDS_acc_no_suffix"].to_list())
    )
    if len(pds_new) > 0:
        for pds in pds_new:
            pds_to_add.append(pds)
        for pds in latest_cluster_center_metadata[
            latest_cluster_center_metadata["PDS_acc_no_suffix"].isin(pds_new)
        ]["asm_acc"].to_list():
            asm_acc_add.append(pds)

    return pds_to_remove, asm_acc_remove, pds_to_add, asm_acc_add


def delete_metadata(conn, asm_acc):
    sql = "DELETE FROM METADATA WHERE asm_acc=?"
    cur = conn.cursor()
    try:
        cur.execute(sql, (asm_acc,))
    except:
        logging.error("Unable to delete %s" % asm_acc)
    cur.execute(sql, (asm_acc,))
    conn.commit()


def update_taxon(args, conn, tmp_folder, database_sig_path):
    c = conn.cursor()
    c.execute("SELECT value FROM DESC where name = 'Species';")
    db_pathogen_name = c.fetchone()[0]
    logging.info("Pathogen name: %s" % db_pathogen_name)
    c.execute("SELECT value FROM DESC where name = 'Version';")
    db_version = c.fetchone()[0]
    logging.info("Pathogen Detection version: %s" % db_version)
    c.execute("SELECT value FROM DESC where name = 'Hash_number';")
    hash_number = int(c.fetchone()[0])
    logging.info("Hash number: %s" % hash_number)
    c.execute("SELECT value FROM DESC where name = 'Kmer_size';")
    kmer_size = int(c.fetchone()[0])
    logging.info("Kmer size: %s" % kmer_size)

    latest_pdg_acc = check_database_version(db_pathogen_name, db_version)

    # get the cluster-centroid table from the original database
    database_cluster_center = pd.read_sql_query("select * from METADATA ", conn)
    # calculate the new cluster-centroid table for the new database
    metadata_file_name, isolate_pds_file = download_metadata(
        db_pathogen_name, latest_pdg_acc, tmp_folder
    )
    latest_metadata = pd.read_csv(
        os.path.join(tmp_folder, metadata_file_name), header=0, sep="\t"
    )
    df_isolate_pds = pd.read_csv(
        os.path.join(tmp_folder, isolate_pds_file), header=0, sep="\t"
    )
    # add cluster accession to metadata
    latest_metadata = latest_metadata.join(
        df_isolate_pds[["target_acc", "PDS_acc"]].set_index("target_acc"),
        on="target_acc",
    )
    latest_metadata_asm = latest_metadata[~latest_metadata["asm_acc"].isnull()]
    calculate_centroid(latest_metadata_asm, latest_pdg_acc, tmp_folder)
    latest_cluster_center = pd.read_csv(
        os.path.join(tmp_folder, f"{latest_pdg_acc}_cluster_center.tsv"), sep="\t"
    )
    latest_cluster_center_metadata = latest_cluster_center.join(
        latest_metadata_asm.drop("PDS_acc", axis=1).set_index("target_acc"),
        on="target_acc",
    )
    # add a new PDS accession column without suffix/PDS version
    database_cluster_center["PDS_acc_no_suffix"] = (
        database_cluster_center["PDS_acc"].str.split(".").str[0]
    )
    latest_cluster_center_metadata["PDS_acc_no_suffix"] = (
        latest_cluster_center_metadata["PDS_acc"].str.split(".").str[0]
    )
    pds_to_remove, asm_acc_remove, pds_to_add, asm_acc_add = compare_tables(
        database_cluster_center, latest_cluster_center_metadata
    )
    logging.info("Found %s PDS to remove" % len(pds_to_remove))
    logging.info("Found %s PDS to add" % len(pds_to_add))

    if len(pds_to_remove) > 0:
        for asm_acc in asm_acc_remove:
            delete_metadata(conn, asm_acc)
    if len(pds_to_add) > 0:
        import_metadata(
            latest_metadata,
            latest_cluster_center_metadata[
                latest_cluster_center_metadata["PDS_acc_no_suffix"].isin(pds_to_add)
            ],
            conn,
        )
        download_and_sketch_assembly(asm_acc_add, hash_number, kmer_size, tmp_folder)

    database_sig = load_file_as_signatures(database_sig_path)
    siglist = []
    for sig in database_sig:
        if str(sig) not in asm_acc_remove:
            siglist.append(sig)
    new_sig_path = os.path.join(tmp_folder, "signature", "*.sig")
    sig_path_list = glob.glob(new_sig_path)
    for sig in sig_path_list:
        siglist.append(load_one_signature(sig))
    with open(database_sig_path, "w") as f:
        save_signatures(siglist, fp=f)
    shutil.rmtree(tmp_folder)

    # update the PDS version info in DESC table in the db
    c.execute(f"Update DESC set value = '{latest_pdg_acc}' where name = 'Version';")

    conn.commit()
    conn.close()
    logging.info(
        f"Database {args.name} updated successfully to version {latest_pdg_acc}"
    )


def update_accession(args, conn, tmp_folder):
    # update the metadata or add new columns
    metadata_file = args.metadata
    if not os.path.exists(metadata_file):
        logging.error("Metadata file not found.")
        exit(1)
    # check if the metadata file has the biosample_acc columns
    metadata = pd.read_csv(metadata_file, header=0)
    if "biosample_acc" not in metadata.columns:
        logging.error("Metadata file does not have biosample_acc column.")
        exit(1)
    # check if other columns are present in the database
    c = conn.cursor()
    c.execute("SELECT * FROM METADATA")
    columns = [description[0] for description in c.description]
    for col in metadata.columns:
        if col not in columns:
            # add the column to the database
            c.execute(f"ALTER TABLE METADATA ADD COLUMN '{col}'")
    conn.commit()
    # import the metadata
    import_metadata(metadata, conn)


def update(args):
    t = time.localtime()
    current_time = time.strftime("%Y%m%d%H%M%S", t)
    print_handler = logging.StreamHandler()
    file_handler = logging.FileHandler("mashpit-" + current_time + ".log")
    print_handler.setLevel(logging.INFO)
    file_handler.setLevel(logging.DEBUG)
    if args.quiet:
        print_handler.setLevel(logging.ERROR)
    logging.basicConfig(
        format="%(asctime)s [%(levelname)s] %(message)s",
        level=logging.DEBUG,
        datefmt="%d-%b-%y %H:%M:%S",
        handlers=[print_handler, file_handler],
    )

    db_folder = os.path.abspath(args.database)
    # create a tmp folder
    cwd = os.getcwd()
    tmp_folder = os.path.join(cwd, "tmp")
    if os.path.isdir(tmp_folder):
        shutil.rmtree(tmp_folder)
    os.mkdir(tmp_folder)
    if not os.path.exists(db_folder):
        logging.error("Path not found.")
        exit(1)
    sql_path = os.path.join(db_folder, args.name + ".db")
    sig_path = os.path.join(db_folder, args.name + ".sig")
    if not (os.path.exists(sql_path)):
        logging.error(f"{sql_path} not found.")
        exit(1)
    if not (os.path.exists(sig_path)):
        logging.error(f"{sig_path} not found.")
        exit(1)

    conn = create_connection(sql_path)
    logging.info("Updating database %s" % args.name)
    c = conn.cursor()
    c.execute("SELECT value FROM DESC where name = 'Type';")
    db_type = c.fetchone()[0]
    logging.info("Database type: %s" % db_type)
    if db_type == "Taxonomy":
        update_taxon(args, conn, tmp_folder, sig_path)
    elif db_type == "Accession":
        update_accession(args, conn, tmp_folder)
